<?php

declare(strict_types=1);

/**
 * SPDX-FileCopyrightText: 2016-2024 Nextcloud GmbH and Nextcloud contributors
 * SPDX-FileCopyrightText: 2015-2016 ownCloud, Inc.
 * SPDX-License-Identifier: AGPL-3.0-only
 */

namespace OCA\Mail\Tests\Unit\Controller;

use ChristophWurst\Nextcloud\Testing\TestCase;
use Horde_Imap_Client_Socket;
use OC\AppFramework\Http\Request;
use OC\Security\CSP\ContentSecurityPolicyNonceManager;
use OCA\Mail\Account;
use OCA\Mail\Attachment;
use OCA\Mail\Contracts\IDkimService;
use OCA\Mail\Contracts\IMailManager;
use OCA\Mail\Contracts\IMailSearch;
use OCA\Mail\Contracts\IMailTransmission;
use OCA\Mail\Contracts\ITrustedSenderService;
use OCA\Mail\Contracts\IUserPreferences;
use OCA\Mail\Controller\MessagesController;
use OCA\Mail\Db\MailAccount;
use OCA\Mail\Db\Tag;
use OCA\Mail\Exception\ClientException;
use OCA\Mail\Exception\ServiceException;
use OCA\Mail\Http\AttachmentDownloadResponse;
use OCA\Mail\Http\HtmlResponse;
use OCA\Mail\IMAP\IMAPClientFactory;
use OCA\Mail\Model\IMAPMessage;
use OCA\Mail\Model\Message;
use OCA\Mail\Service\AccountService;
use OCA\Mail\Service\AiIntegrations\AiIntegrationsService;
use OCA\Mail\Service\ItineraryService;
use OCA\Mail\Service\MailManager;
use OCA\Mail\Service\SmimeService;
use OCA\Mail\Service\SnoozeService;
use OCP\AppFramework\Db\DoesNotExistException;
use OCP\AppFramework\Http;
use OCP\AppFramework\Http\ContentSecurityPolicy;
use OCP\AppFramework\Http\JSONResponse;
use OCP\AppFramework\Http\ZipResponse;
use OCP\AppFramework\Utility\ITimeFactory;
use OCP\Files\Folder;
use OCP\Files\IMimeTypeDetector;
use OCP\IL10N;
use OCP\IRequest;
use OCP\IURLGenerator;
use PHPUnit\Framework\MockObject\MockObject;
use Psr\Log\LoggerInterface;
use ReflectionObject;

class MessagesControllerTest extends TestCase {
	/** @var string */
	private $appName;

	/** @var MockObject|IRequest */
	private $request;

	/** @var MockObject|AccountService */
	private $accountService;

	/** @var MockObject|MailManager */
	private $mailManager;

	/** @var MockObject|IMailSearch */
	private $mailSearch;

	/** @var ItineraryService|MockObject */
	private $itineraryService;

	/** @var string */
	private $userId;

	/** @var MockObject|Folder */
	private $userFolder;

	/** @var MockObject|LoggerInterface */
	private $logger;

	/** @var MockObject|IL10N */
	private $l10n;

	/** @var MessagesController */
	private $controller;

	/** @var MockObject|Account */
	private $account;

	/** @var MockObject|Message */
	private $message;

	/** @var MockObject|Attachment */
	private $attachment;

	/** @var MockObject|IMimeTypeDetector */
	private $mimeTypeDetector;

	/** @var MockObject|IURLGenerator */
	private $urlGenerator;

	/** @var MockObject|ContentSecurityPolicyNonceManager */
	private $nonceManager;

	/** @var MockObject|ITrustedSenderService */
	private $trustedSenderService;

	/** @var MockObject|IMailTransmission */
	private $mailTransmission;

	/** @var ITimeFactory */
	private $oldFactory;

	/** @var MockObject|SmimeService */
	private $smimeService;

	/** @var MockObject|IMAPClientFactory */
	private $clientFactory;
	private IDkimService $dkimService;

	/** @var MockObject|IUserPreferences */
	private $userPreferences;
	private SnoozeService $snoozeService;

	/** @var MockObject|AiIntegrationsService */
	private $aiIntegrationsService;

	protected function setUp(): void {
		parent::setUp();

		$this->appName = 'mail';
		$this->request = $this->getMockBuilder(IRequest::class)->getMock();
		$this->accountService = $this->createMock(AccountService::class);
		$this->mailManager = $this->createMock(IMailManager::class);
		$this->mailSearch = $this->createMock(IMailSearch::class);
		$this->itineraryService = $this->createMock(ItineraryService::class);
		$this->userId = 'john';
		$this->userFolder = $this->createMock(Folder::class);
		$this->request = $this->createMock(Request::class);
		$this->logger = $this->createMock(LoggerInterface::class);
		$this->l10n = $this->createMock(IL10N::class);
		$this->mimeTypeDetector = $this->createMock(IMimeTypeDetector::class);
		$this->urlGenerator = $this->createMock(IURLGenerator::class);
		$this->nonceManager = $this->createMock(ContentSecurityPolicyNonceManager::class);
		$this->trustedSenderService = $this->createMock(ITrustedSenderService::class);
		$this->mailTransmission = $this->createMock(IMailTransmission::class);
		$this->smimeService = $this->createMock(SmimeService::class);
		$this->clientFactory = $this->createMock(IMAPClientFactory::class);
		$this->dkimService = $this->createMock(IDkimService::class);
		$this->userPreferences = $this->createMock(IUserPreferences::class);
		$this->snoozeService = $this->createMock(SnoozeService::class);
		$this->aiIntegrationsService = $this->createMock(AiIntegrationsService::class);

		$timeFactory = $this->createMocK(ITimeFactory::class);
		$timeFactory->expects($this->any())
			->method('getTime')
			->willReturn(10000);
		$this->oldFactory = \OC::$server->offsetGet(ITimeFactory::class);
		\OC::$server->registerService(ITimeFactory::class, function () use ($timeFactory) {
			return $timeFactory;
		});

		$this->controller = new MessagesController(
			$this->appName,
			$this->request,
			$this->accountService,
			$this->mailManager,
			$this->mailSearch,
			$this->itineraryService,
			$this->userId,
			$this->userFolder,
			$this->logger,
			$this->l10n,
			$this->mimeTypeDetector,
			$this->urlGenerator,
			$this->nonceManager,
			$this->trustedSenderService,
			$this->mailTransmission,
			$this->smimeService,
			$this->clientFactory,
			$this->dkimService,
			$this->userPreferences,
			$this->snoozeService,
			$this->aiIntegrationsService,
		);

		$this->account = $this->createMock(Account::class);
		$this->message = $this->createMock(IMAPMessage::class);
		$this->attachment = $this->createMock(Attachment::class);
	}

	protected function tearDown(): void {
		parent::tearDown();

		\OC::$server->offsetUnset(ITimeFactory::class);
		\OC::$server->offsetSet(ITimeFactory::class, $this->oldFactory);
	}

	public function testGetHtmlBody(): void {
		$accountId = 17;
		$mailboxId = 13;
		$folderId = 'testfolder';
		$messageId = 4321;
		$this->account
			->method('getId')
			->willReturn($accountId);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$message = new \OCA\Mail\Db\Message();
		$message->setMailboxId($mailboxId);
		$message->setUid(123);
		$mailbox->setAccountId($accountId);
		$mailbox->setName($folderId);
		$this->mailManager->expects($this->exactly(2))
			->method('getMessage')
			->with($this->userId, $messageId)
			->willReturn($message);
		$this->mailManager->expects($this->exactly(2))
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->exactly(2))
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$client = $this->createMock(Horde_Imap_Client_Socket::class);
		$imapMessage = $this->createMock(IMAPMessage::class);
		$this->mailManager->expects($this->exactly(2))
			->method('getImapMessage')
			->with($client, $this->account, $mailbox, 123, true)
			->willReturn($imapMessage);
		$this->clientFactory->expects($this->exactly(2))
			->method('getClient')
			->with($this->account)
			->willReturn($client);

		$expectedPlainResponse = HtmlResponse::plain('');
		$expectedPlainResponse->cacheFor(3600);

		$nonce = 'abc123';
		$relativeScriptUrl = '/script.js';
		$scriptUrl = 'next.cloud/script.js';
		$this->nonceManager->expects($this->once())
			->method('getNonce')
			->willReturn($nonce);
		$this->urlGenerator->expects($this->once())
			->method('linkTo')
			->with('mail', 'js/htmlresponse.js')
			->willReturn($relativeScriptUrl);
		$this->urlGenerator->expects($this->once())
			->method('getAbsoluteURL')
			->with($relativeScriptUrl)
			->willReturn($scriptUrl);
		$expectedRichResponse = HtmlResponse::withResizer('', $nonce, $scriptUrl);
		$expectedRichResponse->cacheFor(3600);

		$policy = new ContentSecurityPolicy();
		$policy->allowEvalScript(false);
		$policy->disallowScriptDomain('\'self\'');
		$policy->disallowConnectDomain('\'self\'');
		$policy->disallowFontDomain('\'self\'');
		$policy->disallowMediaDomain('\'self\'');
		$expectedPlainResponse->setContentSecurityPolicy($policy);
		$expectedPlainResponse->cacheFor(60 * 60, false, true);
		$expectedRichResponse->setContentSecurityPolicy($policy);
		$expectedRichResponse->cacheFor(60 * 60, false, true);

		$actualPlainResponse = $this->controller->getHtmlBody($messageId, true);
		$actualRichResponse = $this->controller->getHtmlBody($messageId, false);

		$this->assertEquals($expectedPlainResponse, $actualPlainResponse);
		$this->assertEquals($expectedRichResponse, $actualRichResponse);
	}

	public function testDownloadAttachment() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$uid = 321;
		$attachmentId = '3';

		// Attachment data
		$contents = 'abcdef';
		$name = 'cat.jpg';
		$type = 'image/jpg';
		$message = new \OCA\Mail\Db\Message();
		$message->setMailboxId($mailboxId);
		$message->setUid($uid);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->mailManager->expects($this->once())
			->method('getMailAttachment')
			->with($this->account, $mailbox, $message, $attachmentId)
			->will($this->returnValue($this->attachment));
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$this->attachment->expects($this->once())
			->method('getContent')
			->will($this->returnValue($contents));
		$this->attachment->expects($this->any())
			->method('getName')
			->will($this->returnValue($name));
		$this->attachment->expects($this->once())
			->method('getType')
			->will($this->returnValue($type));

		$expected = new AttachmentDownloadResponse($contents, $name, $type);
		$response = $this->controller->downloadAttachment(
			$id,
			$attachmentId
		);

		$this->assertEquals($expected, $response);
	}

	public function testSaveSingleAttachment() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$uid = 321;
		$attachmentId = '2.2';
		$targetPath = 'Downloads';
		$message = new \OCA\Mail\Db\Message();
		$message->setMailboxId($mailboxId);
		$message->setUid($uid);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$this->mailManager->expects($this->once())
			->method('getMailAttachment')
			->with($this->account, $mailbox, $message, $attachmentId)
			->will($this->returnValue($this->attachment));
		$this->attachment->expects($this->once())
			->method('getName')
			->with()
			->will($this->returnValue('cat.jpg'));
		$this->userFolder->expects($this->once())
			->method('nodeExists')
			->with('Downloads/cat.jpg')
			->will($this->returnValue(false));
		$file = $this->getMockBuilder('\OCP\Files\File')
			->disableOriginalConstructor()
			->getMock();
		$this->userFolder->expects($this->once())
			->method('newFile')
			->with('Downloads/cat.jpg')
			->will($this->returnValue($file));
		$file->expects($this->once())
			->method('putContent')
			->with('abcdefg');
		$this->attachment->expects($this->once())
			->method('getContent')
			->will($this->returnValue('abcdefg'));

		$expected = new JSONResponse();
		$response = $this->controller->saveAttachment(
			$id,
			$attachmentId,
			$targetPath
		);

		$this->assertEquals($expected, $response);
	}

	public function testSaveAllAttachments() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$uid = 321;
		$attachmentId = '0';
		$targetPath = 'Downloads';
		$message = new \OCA\Mail\Db\Message();
		$message->setMailboxId($mailboxId);
		$message->setUid($uid);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$client = $this->createMock(Horde_Imap_Client_Socket::class);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));

		$this->mailManager->expects($this->once())
			->method('getMailAttachments')
			->with($this->account, $mailbox, $message)
			->will($this->returnValue([$this->attachment]));
		$this->attachment->expects($this->once())
			->method('getName')
			->with()
			->will($this->returnValue('cat.jpg'));
		$this->userFolder->expects($this->once())
			->method('nodeExists')
			->with('Downloads/cat.jpg')
			->will($this->returnValue(false));
		$file = $this->getMockBuilder('\OCP\Files\File')
			->disableOriginalConstructor()
			->getMock();
		$this->userFolder->expects($this->once())
			->method('newFile')
			->with('Downloads/cat.jpg')
			->will($this->returnValue($file));
		$file->expects($this->once())
			->method('putContent')
			->with('abcdefg');
		$this->attachment->expects($this->once())
			->method('getContent')
			->will($this->returnValue('abcdefg'));

		$expected = new JSONResponse();
		$response = $this->controller->saveAttachment(
			$id,
			$attachmentId,
			$targetPath
		);

		$this->assertEquals($expected, $response);
	}

	public function testDownloadAttachments() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$uid = 321;
		$message = new \OCA\Mail\Db\Message();
		$message->setMailboxId($mailboxId);
		$message->setUid($uid);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$attachments = [
			new Attachment(
				null,
				'cat.png',
				'image/png',
				'abcdefg',
				7,
			),
		];

		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));

		//
		$this->mailManager->expects($this->once())
			->method('getMailAttachments')
			->with($this->account, $mailbox, $message)
			->willReturn($attachments);
		// build our zip
		$response = $this->controller->downloadAttachments(
			$id
		);

		$this->assertInstanceOf(ZipResponse::class, $response);

		$zip = new ZipResponse($this->request, 'attachments');
		foreach ($attachments as $attachment) {
			$fileName = $attachment->getName();
			$fh = fopen('php://temp', 'r+');
			fputs($fh, $attachment->getContent());
			$size = $attachment->getSize();
			rewind($fh);
			$zip->addResource($fh, $fileName, $size);
		}

		// Reflection is needed to get private properties
		$refZip = new ReflectionObject($zip);
		$prop = $refZip->getProperty('resources');
		$prop->setAccessible(true);
		$zipValues = $prop->getValue($zip);
		$refResponse = new ReflectionObject($response);
		$prop = $refResponse->getProperty('resources');
		$prop->setAccessible(true);
		$responseValues = $prop->getValue($zip);

		$this->assertTrue(is_resource($zipValues[0]['resource']));
		$this->assertTrue(is_resource($responseValues[0]['resource']));

		// ZipResponse will write fopen id into the array
		// so assert equals needs to have these values unset before comparison
		unset($zipValues[0]['resource']);
		unset($responseValues[0]['resource']);

		$this->assertEquals($zipValues, $responseValues);
	}

	public function testDownloadAttachmentsNoAccountError() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$uid = 321;
		$message = new \OCA\Mail\Db\Message();
		$message->setMailboxId($mailboxId);
		$message->setUid($uid);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->willThrowException(new ClientException());


		// test our json error response
		$this->expectException(ClientException::class);
		$response = $this->controller->downloadAttachments(
			$id
		);

		$this->assertInstanceOf(JSONResponse::class, $response);
	}

	public function testDownloadAttachmentsNoMailboxError() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$uid = 321;
		$message = new \OCA\Mail\Db\Message();
		$message->setMailboxId($mailboxId);
		$message->setUid($uid);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);

		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->willThrowException(new ClientException());

		// test our json error response
		$this->expectException(ClientException::class);
		$response = $this->controller->downloadAttachments(
			$id
		);

		$this->assertInstanceOf(JSONResponse::class, $response);
	}

	public function testDownloadAttachmentsNoMessageError() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$uid = 321;
		$message = new \OCA\Mail\Db\Message();
		$message->setMailboxId($mailboxId);
		$message->setUid($uid);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);

		$this->mailManager->expects($this->once())
			->method('getMessage')
			->willThrowException(new ServiceException());
		// test our json error response
		$this->expectException(ServiceException::class);
		$response = $this->controller->downloadAttachments(
			$id
		);

		$this->assertInstanceOf(JSONResponse::class, $response);
	}


	public function testSetFlagsUnseen() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$flags = [
			'seen' => false
		];
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$this->mailManager->expects($this->once())
			->method('flagMessage')
			->with($this->account, 'INBOX', 444, 'seen', false);

		$expected = new JSONResponse();
		$response = $this->controller->setFlags(
			$id,
			$flags
		);

		$this->assertEquals($expected, $response);
	}

	public function testSetTagFailing() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 1;
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$message->setMessageId('<jhfjkhdsjkfhdsjkhfjkdsh@test.com>');
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->willThrowException(new DoesNotExistException(''));
		$this->mailManager->expects($this->never())
			->method('getTagByImapLabel');
		$this->mailManager->expects($this->never())
			->method('tagMessage');

		$this->controller->setTag($id, Tag::LABEL_IMPORTANT);
	}

	public function testSetTagNotFound() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 1;
		$imapLabel = '$label6';
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$message->setMessageId('<jhfjkhdsjkfhdsjkhfjkdsh@test.com>');
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$this->mailManager->expects($this->once())
			->method('getTagByImapLabel')
			->with($imapLabel, $this->userId)
			->willThrowException(new ClientException('Computer says no'));
		$this->mailManager->expects($this->never())
			->method('tagMessage');

		$this->controller->setTag($id, $imapLabel);
	}

	public function testSetTag() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 1;
		$tag = new Tag();
		$tag->setImapLabel(Tag::LABEL_IMPORTANT);
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$message->setMessageId('<jhfjkhdsjkfhdsjkhfjkdsh@test.com>');
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$this->mailManager->expects($this->once())
			->method('getTagByImapLabel')
			->with($tag->getImapLabel(), $this->userId)
			->willReturn($tag);
		$this->mailManager->expects($this->once())
			->method('tagMessage')
			->with($this->account, $mailbox->getName(), $message, $tag, true);

		$this->controller->setTag($id, $tag->getImapLabel());
	}

	public function testRemoveTagFailing() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 1;
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$message->setMessageId('<jhfjkhdsjkfhdsjkhfjkdsh@test.com>');
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->willThrowException(new DoesNotExistException(''));
		$this->mailManager->expects($this->never())
			->method('getTagByImapLabel');
		$this->mailManager->expects($this->never())
			->method('tagMessage');

		$this->controller->removeTag($id, Tag::LABEL_IMPORTANT);
	}

	public function testRemoveTagNotFound() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 1;
		$imapLabel = '$label6';
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$message->setMessageId('<jhfjkhdsjkfhdsjkhfjkdsh@test.com>');
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$this->mailManager->expects($this->once())
			->method('getTagByImapLabel')
			->with($imapLabel, $this->userId)
			->willThrowException(new ClientException('Computer says no'));
		$this->mailManager->expects($this->never())
			->method('tagMessage');

		$this->controller->removeTag($id, $imapLabel);
	}

	public function testRemoveTag() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 1;
		$tag = new Tag();
		$tag->setImapLabel(Tag::LABEL_IMPORTANT);
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$message->setMessageId('<jhfjkhdsjkfhdsjkhfjkdsh@test.com>');
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$this->mailManager->expects($this->once())
			->method('getTagByImapLabel')
			->with($tag->getImapLabel(), $this->userId)
			->willReturn($tag);
		$this->mailManager->expects($this->once())
			->method('tagMessage')
			->with($this->account, $mailbox->getName(), $message, $tag, false);

		$this->controller->removeTag($id, $tag->getImapLabel());
	}

	public function testSetFlagsFlagged() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$flags = [
			'flagged' => true
		];
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$this->mailManager->expects($this->once())
			->method('flagMessage')
			->with($this->account, 'INBOX', 444, 'flagged', true);

		$expected = new JSONResponse();
		$response = $this->controller->setFlags(
			$id,
			$flags
		);

		$this->assertEquals($expected, $response);
	}

	public function testDestroy() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$this->mailManager->expects($this->once())
			->method('deleteMessage')
			->with($this->account, 'INBOX', 444);

		$expected = new JSONResponse();
		$result = $this->controller->destroy($id);

		$this->assertEquals($expected, $result);
	}

	public function testDestroyWithAccountNotFound() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->throwException(new DoesNotExistException('')));

		$expected = new JSONResponse([], Http::STATUS_FORBIDDEN);

		$this->assertEquals($expected, $this->controller->destroy($id));
	}

	public function testDestroyWithFolderOrMessageNotFound() {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);
		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$this->mailManager->expects($this->once())
			->method('deleteMessage')
			->with($this->account, 'INBOX', 444)
			->willThrowException(new ServiceException());
		$this->expectException(ServiceException::class);

		$this->controller->destroy($id);
	}

	public function testGetThread(): void {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$message->setThreadRootId('<marlon@slimehunter.com>');
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);

		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->willReturn($this->account);
		$this->mailManager->expects($this->once())
			->method('getThread')
			->with($this->account, $message->getThreadRootId());

		$this->controller->getThread($id);
	}

	public function testGetThreadNoThreadRootId(): void {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);

		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->willReturn($this->account);
		$this->mailManager->expects($this->never())
			->method('getThread');

		$this->controller->getThread($id);
	}

	public function testGetThreadThreadRootIdEmptyString(): void {
		$accountId = 17;
		$mailboxId = 987;
		$id = 123;
		$message = new \OCA\Mail\Db\Message();
		$message->setUid(444);
		$message->setMailboxId($mailboxId);
		$message->setMessageId('<123@cde.com>');
		$message->setThreadRootId('');
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setName('INBOX');
		$mailbox->setAccountId($accountId);

		$this->mailManager->expects($this->once())
			->method('getMessage')
			->with($this->userId, $id)
			->willReturn($message);
		$this->mailManager->expects($this->once())
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->once())
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->willReturn($this->account);
		$this->mailManager->expects($this->once())
			->method('getThread');

		$this->controller->getThread($id);
	}

	public function testExport() {
		$accountId = 17;
		$mailboxId = 13;
		$folderId = 'testfolder';
		$messageId = 4321;
		$this->account
			->method('getId')
			->willReturn($accountId);
		$mailbox = new \OCA\Mail\Db\Mailbox();
		$message = new \OCA\Mail\Db\Message();
		$message->setMailboxId($mailboxId);
		$message->setUid(123);
		$message->setSubject('core/master has new results');
		$mailbox->setAccountId($accountId);
		$mailbox->setName($folderId);
		$this->mailManager->expects($this->exactly(1))
			->method('getMessage')
			->with($this->userId, $messageId)
			->willReturn($message);
		$this->mailManager->expects($this->exactly(1))
			->method('getMailbox')
			->with($this->userId, $mailboxId)
			->willReturn($mailbox);
		$this->accountService->expects($this->exactly(1))
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($accountId))
			->will($this->returnValue($this->account));
		$source = file_get_contents(__DIR__ . '/../../data/mail-message-123.txt');
		$client = $this->createMock(Horde_Imap_Client_Socket::class);
		$this->mailManager->expects($this->exactly(1))
			->method('getSource')
			->with($client, $this->account, $folderId, 123)
			->willReturn($source);
		$this->clientFactory->expects($this->once())
			->method('getClient')
			->willReturn($client);

		$expectedResponse = new AttachmentDownloadResponse(
			$source,
			'core/master has new results.eml',
			'message/rfc822'
		);
		$actualResponse = $this->controller->export($messageId);

		$this->assertEquals($expectedResponse, $actualResponse);
	}

	public function testGetDkim() {
		$mailAccount = new MailAccount();
		$mailAccount->setId(100);
		$mailAccount->setUserId($this->userId);
		$account = new Account($mailAccount);

		$mailbox = new \OCA\Mail\Db\Mailbox();
		$mailbox->setId(4);
		$mailbox->setAccountId($account->getId());
		$mailbox->setName('FooBar');

		$message = new \OCA\Mail\Db\Message();
		$message->setId(4448);
		$message->setMailboxId($mailbox->getId());
		$message->setUid(123);
		$message->setSubject('core/master has new results');


		$this->mailManager->expects($this->exactly(1))
			->method('getMessage')
			->with($this->userId, $message->getId())
			->willReturn($message);
		$this->mailManager->expects($this->exactly(1))
			->method('getMailbox')
			->with($this->userId, $mailbox->getId())
			->willReturn($mailbox);
		$this->accountService->expects($this->exactly(1))
			->method('find')
			->with($this->equalTo($this->userId), $this->equalTo($account->getId()))
			->will($this->returnValue($account));
		$this->dkimService->expects($this->exactly(1))
			->method('validate')
			->with($account, $mailbox, $message->getUid())
			->willReturn(true);

		$actualResponse = $this->controller->getDkim($message->getId());

		$this->assertInstanceOf(JSONResponse::class, $actualResponse);
		$this->assertEquals(['valid' => true], $actualResponse->getData());
	}
}
